// viterbi_benchmark




////////////////////////////////////////////
#define TRIALS  1000      // Frames      //
#define TESTLOOPS 10000 // Speedtest    //
/////////////////////////////////////////

#undef UNICODE
#define _CRT_SECURE_NO_WARNINGS // We don't fly to the moon..

#include <windows.h>
#include <stdio.h>
#include <math.h>


#define MALLOC(a) _aligned_malloc(a, 64)
#define FREE(a) _aligned_free(a)
#define COMPUTETYPE unsigned int
#define K 7
#define RATE 4
#define PLI (LARGE_INTEGER*)
#define MAXFRAMEBITS (24 * 384) // max bitrate 384 kBit/s for old DAB
#define bitlen ((MAXFRAMEBITS + (K - 1)) / 8 + 1)
#define OFFSET (127.5)
#define CLIP 255
#define EBN0 (3.0)
double Gain = 32.0;
#define	random() rand()
#define	srandom(time) srand(time)
#define POLYS { 109, 79, 83, 109 }



char szReferenceDll[] = "viterbi_qirx.dll";
char szDllName[] = "viterbi.dll";
char szIniName[] = "viterbi.txt";

char szFnDeconvolve[] = "deconvolve";


typedef int (WINAPI* DECONVOLVE)(int frameBits, COMPUTETYPE* symbols,
    int, unsigned char* decodedBits);
typedef int (WINAPI* GETCPUCAPS)();
typedef int (WINAPI* INITIALIZE)();

INITIALIZE initialize;
DECONVOLVE deconvolve, deconReference;
GETCPUCAPS GetCPUCaps;

// Bitmasks returned by GetCPUCaps
#define CpuHasSSE2  0b00000000
#define CpuHasSSE3  0b00000001
#define CpuHasSSSE3 0b00000010
#define CpuHasSSE41 0b00000100
#define CpuHasSSE42 0b00001000
#define CpuHasAVX1  0b00010000
#define CpuHasAVX2  0b00100000
#define CpuHasFMA3  0b01000000
// CpuHasAVX5 is only set if the CPU supports AVX512F+BW+VL!
#define CpuHasAVX5  0b10000000


// popcount algo taken from "Hackers Delight" to avoid problems
// with CPU's w/o popcount-instruction.
int popcount32(unsigned int x) {
    unsigned int n;
    n = (x >> 1) & 0x77777777; // Count bits in
    x = x - n; // each 4-bit
    n = (n >> 1) & 0x77777777; // field.
    x = x - n;
    n = (n >> 1) & 0x77777777;
    x = x - n;
    x = (x + (x >> 4)) & 0x0F0F0F0F; // Get byte sums.
    x = x * 0x01010101; // Add the bytes.
    return x >> 24;
}
// Determine parity of argument: 1 = odd, 0 = even
static inline int parityb(unsigned char x)
{
    __asm__ __volatile__("test %1,%1;setpo %0" : "=g"(x) : "r" (x));
    return x;
}
// Generate gaussian random double with specified mean and std_dev
static inline double normal_rand(double mean, double std_dev)
{
    double fac,rsq,v1,v2;
    static double gset;
    static int iset;

    if(iset)
    {
        iset = 0;
        return mean + std_dev*gset;
    }

    do
    {
        v1 = 2.0 * (double)random() / RAND_MAX - 1;
        v2 = 2.0 * (double)random() / RAND_MAX - 1;
        rsq = v1*v1 + v2*v2;
    } while(rsq >= 1.0 || rsq == 0.0);
    fac = sqrt(-2.0*log(rsq)/rsq);
    gset = v1*fac;
    iset++;
    return mean + std_dev*v2*fac;
}
static inline unsigned char addnoise(int sym,double amp,double gain,
    double offset,int clip)
{
    int sample;

    sample = offset + gain*normal_rand(sym?amp:-amp,1.0);

    if(sample < 0)
        sample = 0;
    else if(sample > clip)
        sample = clip;
    return sample;
}


int GetProcessorInfo(char* szCpuID, char* szCpuFeat);
int GetUserLocalAppDataFolder(char* szPath);
int SetIniFileValue(char* szFullIniFilePath, int val);
int GetIniFileValue(char* szFullIniFilePath);

const char* FuncTable[] = {
    "SSE2",
    "SSSE3",
    "SSE4.1",
    "AVX",
    "AVX2",
    "AVX512F"
};


int main() {

    HMODULE hViterbi, hVitRef;
    COMPUTETYPE* inputInts;
    unsigned char *originalBits, *decodedBits;
    double dQpfPeriod, gain, esn0, execTime;
    LARGE_INTEGER qpf, start, finish;
    long long compval, compare[16];
    int i, j, tr, bitrate, framebits, Len, sr, errcnt, badframes,
        tot_errs, iFnNum, iCpuCaps, iPathStatus;
    int oldIniValue = -1; // assume automatic
    int polys[RATE] = POLYS;
    char szFullIniPath[260], szTmp[260], szCpuID[256], szCpuFeat[128];;

    QueryPerformanceFrequency(&qpf);
    dQpfPeriod = 1.0 / qpf.QuadPart;

// Wait a little bit until the console-window appeared on screen.
// Otherwise the desktop refreshes and the info-rect from the
// viterbi.dll will be erased after some millisecs.
    Sleep(1000); 



#if (1)

    printf("viterbi - test & benchmark\n");

// Try to load the viterbi.dll and get the exported proc-addresses
    hViterbi = LoadLibrary(szDllName);
    if (!hViterbi) {
        printf("Could't load %s!\n", szDllName);
        goto NoViterbi;
    }

    initialize = (INITIALIZE)GetProcAddress(hViterbi, "initialize");
    if (!initialize) {
        printf("Entrypoint 'initialize' not found.\n");
        goto NoViterbi;
    }
    
    deconvolve = (DECONVOLVE)GetProcAddress(hViterbi, szFnDeconvolve);
    if (!deconvolve) {
        printf("Entrypoint '%s' not found.\n", szFnDeconvolve);
        goto NoViterbi;
    }

    GetCPUCaps = (GETCPUCAPS)GetProcAddress(hViterbi, "GetCPUCaps");
    if (!GetCPUCaps) {
        printf("Entrypoint 'GetCPUCaps' not found.\n");
        goto NoViterbi;
    }

// Show at least the CPU-Info
    iCpuCaps = GetProcessorInfo(szCpuID, szCpuFeat);
    printf("\n%s\n%s\n\n", szCpuID, szCpuFeat);


// Try to load Reference-DLL
    hVitRef = LoadLibrary(szReferenceDll);
    if (!hVitRef) {
        printf("Could't load reference %s!\n", szReferenceDll);
        goto NoViterbiReference;
    }

    deconReference = (DECONVOLVE)GetProcAddress(hVitRef, szFnDeconvolve);
    if (!deconReference) {
        printf("Entrypoint '%s' not found.\n", szFnDeconvolve);
        goto NoViterbiReference;
    }

// Get the memory for the arrays
    inputInts = (COMPUTETYPE*)MALLOC(RATE * (MAXFRAMEBITS + (K - 1)) * sizeof(COMPUTETYPE));
    if (!inputInts) {
        printf("Allocation of inputInts array failed\n");
        goto NoMem;
    }

    originalBits = (unsigned char*)MALLOC(bitlen);
    if (!originalBits) {
        printf("Allocation of originalBits array failed\n");
        FREE(inputInts);
        goto NoMem;
    }

    decodedBits = (unsigned char*)MALLOC(bitlen);
    if (!decodedBits) {
        printf("Allocation of decodedBits array failed\n\n");
        FREE(originalBits);
        FREE(inputInts);
        goto NoMem;
    }


// Search the inifile
    iPathStatus = GetUserLocalAppDataFolder(szFullIniPath);

    if (2 == iPathStatus) {
        sprintf(szTmp, "%s\\%s", szFullIniPath, szIniName);
        strcpy(szFullIniPath, szTmp);
        oldIniValue = GetIniFileValue(szFullIniPath);
        if (oldIniValue != -2)
            iPathStatus = 1; // inifile exists
    }
    else
        iPathStatus = 0; // no inifile 

    if (!iPathStatus) // unlikely
        printf("Inifile of %s not found! The DLL uses automatic CPU-dispatching!\n\n", szDllName);

// Boost
    SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS);
    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);

// test & benchmark the reference DLL
    printf("Testing function \"%s\" from %s with %d calls...\n",
        szFnDeconvolve, szReferenceDll, TESTLOOPS);


    esn0 = EBN0 + 10 * log10(1. / ((double)RATE));
    gain = 1. / sqrt(0.5 / pow(10., esn0 / 10.));

    bitrate = 128;
    framebits = bitrate * 24;
    Len = ((framebits + (K - 1)) * RATE);
    tot_errs = 0;
    badframes = 0;
    sr = 0;
    srandom(0);

    for (tr = 0; tr < TRIALS; tr++)
    {
        for (i = 0; i < framebits + (K - 1); i++)
        {
            int bit = (i < framebits) ? (random() & 1) : 0;
            sr = (sr << 1) | bit;
            originalBits[i / 8] = sr & 0xff;
            for (j = 0; j < RATE; j++)
                inputInts[RATE * i + j] = addnoise(parityb(sr & polys[j]), gain,
                    Gain, OFFSET, CLIP);
        }

        deconReference(framebits, inputInts, Len, decodedBits);

        errcnt = 0;
        //_asm { int 3};
        for (int i = 0; i < framebits / 8; i++)
        {
            int e = popcount32(decodedBits[i] ^ originalBits[i]);
            errcnt += e;
            tot_errs += e;
        }
        if (errcnt != 0)
            badframes++;

    }
    printf("BER %d/%d (%10.3g) FER %d/%d (%10.3g)\n",
        tot_errs, framebits * (tr), tot_errs / ((double)framebits * (tr)),
        badframes, tr, (double)badframes / (tr));


    j = 0;
    for (bitrate = 32; bitrate < 160; bitrate += 32)
    {
        framebits = bitrate * 24;
        Len = ((framebits + (K - 1)) * RATE);
        printf("Bitrate: %5d\tFramebits: %5d\t", bitrate, framebits);

        for (tr = 0; tr < TESTLOOPS / 4; tr++) // warm up
            deconReference(framebits, inputInts, Len, decodedBits);
//        SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
        QueryPerformanceCounter(PLI & start);
        for (tr = 0; tr < TESTLOOPS; tr++)
            deconReference(framebits, inputInts, Len, decodedBits);
        QueryPerformanceCounter(PLI & finish);

        compare[j] = finish.QuadPart - start.QuadPart;
        execTime = compare[j] * dQpfPeriod;
        printf("Time: %8.4f sec\n", execTime);
        j++;
    }

#endif
     
    printf("\n-------------------------------------------------------------------------------\n");

#if (1)
    for (iFnNum = 0; iFnNum < 6; iFnNum++) {
        if (((iFnNum == 1) && !(iCpuCaps & CpuHasSSSE3)) ||
            ((iFnNum == 2) && !(iCpuCaps & CpuHasSSE41)) ||
            ((iFnNum == 3) && !(iCpuCaps & CpuHasAVX1)) ||
            ((iFnNum == 4) && !(iCpuCaps & CpuHasAVX2)) ||
            ((iFnNum == 5) && !(iCpuCaps & CpuHasAVX5)))
            continue; // skip unsupported instruction sets

        if (iPathStatus) {
            SetIniFileValue(szFullIniPath, iFnNum);
            initialize(); // updating the CPU-dispatcher
            printf("\nTesting function \"%s\" (%s) from %s with %d calls...\n",
                szFnDeconvolve, FuncTable[iFnNum], szDllName, TESTLOOPS);
        }
        else {
            printf("\nTesting function \"%s\" (auto) from %s with %d calls...\n",
                szFnDeconvolve, szDllName, TESTLOOPS);
        }

        bitrate = 128;
        framebits = bitrate * 24;
        Len = ((framebits + (K - 1)) * RATE);
        tot_errs = 0;
        badframes = 0;
        sr = 0;
        srandom(0);

        for (tr = 0; tr < TRIALS; tr++) {

            for (i = 0; i < framebits + (K - 1); i++) {
                int bit = (i < framebits) ? (random() & 1) : 0;
                sr = (sr << 1) | bit;
                originalBits[i / 8] = sr & 0xff;
                for (j = 0; j < RATE; j++)
                    inputInts[RATE * i + j] = addnoise(parityb(sr & polys[j]),
                        gain, Gain, OFFSET, CLIP);
            }

            deconvolve(framebits, inputInts, Len, decodedBits);
            errcnt = 0;

            for (int i = 0; i < framebits / 8; i++) {
                int e = popcount32(decodedBits[i] ^ originalBits[i]);
                errcnt += e;
                tot_errs += e;
            }

            if (errcnt != 0)
                badframes++;
        }
        printf("BER %d/%d (%10.3g) FER %d/%d (%10.3g)\n",
            tot_errs, framebits * (tr), tot_errs / ((double)framebits * (tr)),
            badframes, tr, (double)badframes / (tr));

        for (bitrate = 32, j = 0; bitrate < 160; bitrate += 32, j++) {
            framebits = bitrate * 24;
            Len = ((framebits + (K - 1)) * RATE);
            printf("Bitrate: %5d\tFramebits: %5d\t", bitrate, framebits);

            for (tr = 0; tr < TESTLOOPS / 4; tr++) // warm up
                deconvolve(framebits, inputInts, Len, decodedBits);
//            SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
            QueryPerformanceCounter(PLI & start);
            for (tr = 0; tr < TESTLOOPS; tr++)
                deconvolve(framebits, inputInts, Len, decodedBits);
            QueryPerformanceCounter(PLI & finish);

            compval = finish.QuadPart - start.QuadPart;
            execTime = compval * dQpfPeriod;
            printf("Time: %8.4f sec", execTime);
            if (compval < compare[j]) {
                execTime = compare[j] / (double)compval;
                printf("\t%6.3f x FASTER\n", execTime);
            }
            else
                printf("\n");
        }
        if (!iPathStatus) // exit the loop if inifile was not found
            break;
    }
    
    if(iPathStatus) // restore original value
        SetIniFileValue(szFullIniPath, oldIniValue);
#endif

    FREE(originalBits);
    FREE(decodedBits);
    FREE(inputInts);
NoMem:
    FreeLibrary(hVitRef);
NoViterbiReference:
    FreeLibrary(hViterbi);
NoViterbi:
    printf("\n\nPress a key\n");
    getchar();
    return 0;
}




int GetProcessorInfo(char* szCpuID, char* szCpuFeat) {

    char tmp[256];
    int retVal = GetCPUCaps(); 
    *szCpuFeat = '\0';
    GetEnvironmentVariable("PROCESSOR_IDENTIFIER", szCpuID, 256);
    strcat(szCpuID, "\nCPU Level: ");
    GetEnvironmentVariable("PROCESSOR_LEVEL", tmp, 256);
    strcat(szCpuID, tmp);
    strcat(szCpuID, " / Revision: ");
    GetEnvironmentVariable("PROCESSOR_REVISION", tmp, 256);
    strcat(szCpuID, tmp);
    
    strcat(szCpuFeat, "Instruction sets: SSE2");

    if (retVal & CpuHasSSE3)
        strcat(szCpuFeat, ", SSE3");
    if (retVal & CpuHasSSSE3)
        strcat(szCpuFeat, ", SSSE3");
    if (retVal & CpuHasSSE41)
        strcat(szCpuFeat, ", SSE4.1");
    if (retVal & CpuHasSSE42)
        strcat(szCpuFeat, ", SSE4.2");
    if (retVal & CpuHasAVX1)
        strcat(szCpuFeat, ", AVX");
    if (retVal & CpuHasAVX2)
        strcat(szCpuFeat, ", AVX2");
    if (retVal & CpuHasFMA3)
        strcat(szCpuFeat, ", FMA3");
    if (retVal & CpuHasAVX5)
        strcat(szCpuFeat, ", AVX512F+BW+VL");

    return retVal;
}

// GetUserLocalAppDataFolder return values:
// 0 : env-variable not set, path too long or a file "viterbi" exists
//     in the "...\AppData\Local\" directory for unknown reasons
// 1 : the path "...\AppData\Local\" exists
// 2 : the path "...\AppData\Local\viterbi\" exists
// taken from the viterbi.dll sources
int GetUserLocalAppDataFolder(char* szLocalAppDataPath) {
    char tempbuf[MAX_PATH];
    WIN32_FIND_DATA wfd;
    HANDLE hSearch;
    int PathLen, iPathStatus = 0; // assume error

    // If GetEnvironmentVariable() fails or if the current path is too long
    // for appending our directory and the filename ("\viterbi\viterbi.txt")
    // later, we'll better fail at this point.
    PathLen = GetEnvironmentVariable("LOCALAPPDATA", tempbuf, MAX_PATH - 21);

    if (!PathLen || PathLen >= MAX_PATH - 21) {
        *szLocalAppDataPath = 0;
        return iPathStatus;
    }

    // Otherwise we returning at least the "AppData\Local" path, so our own
    // directory can be created later w/o asking for the path again.
    strcpy(szLocalAppDataPath, tempbuf);
    iPathStatus = 1;

    // Append our own directory-name to the path and search for it
    sprintf(tempbuf, "%s\\%s", szLocalAppDataPath, "viterbi");
    hSearch = FindFirstFile(tempbuf, &wfd);

    if (INVALID_HANDLE_VALUE != hSearch) {
        // We have a match, but this can be a file. This is unlikely, but
        // who knows? Let's check it.

        if (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
            // It's a directory...
            strcpy(szLocalAppDataPath, tempbuf); // copy the path
            iPathStatus = 2; // all is fine
            SetFileAttributes(tempbuf, FILE_ATTRIBUTE_NORMAL);
        }
        FindClose(hSearch);
    }
    return iPathStatus;
}

// Sets the "instruction set override" value (first byte in viterbi.txt)
int SetIniFileValue(char* szFullIniFilePath, int val) {
    char fib = '0';
    int retVal = 0;
    if (val > 5) return -1;

    FILE* fini = fopen(szFullIniFilePath, "rb+");
    if (fini) {

        if (val >= 0)
            fib += char(val);
        else
            fib = 'x';

        fwrite(&fib, 1, 1, fini);
        fclose(fini);
        retVal = 1;
    }
    return retVal;
}
// Gets the instruction set value from viterbi.txt.
// Returns -2 if the file was not found
int GetIniFileValue(char* szFullIniFilePath) {
    char fib = 0;
    int retVal = -2; // reval if file not found
    FILE* fini = fopen(szFullIniFilePath, "rb+");
    if (fini) {

        fread(&fib, 1, 1, fini);
        fclose(fini);
        if (fib >= '0' && fib < '6')
            retVal = fib - '0';   // 0 to 5
        else
            retVal = -1; // automatic
    }
    return retVal;
}
