/*
* This file is part of
*
* viterbi.dll replacement for QIRX-SDR (Windows 64 Bit)
*
* 
* This program is free software; you can redistribute it and/or modify it
* under the terms of the GNU General Public License as published by the Free
* Software Foundation; either version 2 of the License, or (at your option)
* any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
* or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
* more details.
*
* You should have received a copy of the GNU General Public License along with
* this program; if not, write to the Free Software Foundation, Inc., 59 Temple
* Place - Suite 330, Boston, MA 02111-1307, USA
* and it is distributed under the GNU General Public License (GPL).
*
* (c) 2021-2023 Heiko Vogel <hevog@gmx.de>
*/

#define ALIGN(a) __declspec(align(a))
#define align(a) ALIGN(a)

// If VIT_WRITE_LOGFILE is defined, the DLL will be compiled with code
// that writes a lot of information about the viterbi-decoder and the
// RS-decoder into a file. This logfile contains, among some other things,
// timestamps, ThreadIDs and the addresses of the buffers transferred by
// QIRX. The file is located at 
//           ...\APPDATA\LOCAL\viterbi\YYYYMMDD_HHMMSS.log
// and its size increases at a rate of approximately 65 MB/hr, so don't
// forget to swap this build of the DLL back to the production DLL after 
// your investigations.
//#define VIT_WRITE_LOGFILE

// Includes information about the creation and termination of threads.
//#define VIT_INCLUDE_THREAD_ATTACH_DETACH

// Includes a short summary at the end of the logfile.
#define VIT_INCLUDE_SUMMARY


// If VIT_WRITE_LOGFILE is defined you can define VIT_WRITE_SYMBOLS too.
// This creates a second file in which the symbols for the viterbi-decoder
// are stored in binary form (char). The file name is the same like above
// but has the extension *.sym. This file grows real fast!
//#define VIT_WRITE_SYMBOLS


// RENORMALIZE_THRESHOLD for the viterbi-decoders in 'deconvolve.cpp'.
// Use RENORMALIZE_THRESHOLD in 'const.inc' for the assembly-language parts.
// Keep the value below 180.

#define RENORMALIZE_THRESHOLD 150


// Some well known defines for the viterbi decoder.
#define K 7
#define NUMSTATES  64
#define COMPUTETYPE  unsigned char
#define DECISIONTYPE unsigned char
#define DECISIONTYPE_BITSIZE 8

typedef union {
    DECISIONTYPE t[NUMSTATES / DECISIONTYPE_BITSIZE];
    unsigned int w[NUMSTATES / 32];
} decision_t;


// Some defines for the Reed-Solomon decoder
#define c_nn 255
#define c_gfpoly 285
#define c_nroots 10




// RSLookup allocated once at startup in DllMain. The members are heavily used
// in RScheckSuperframe. If 'LOCAL_COPY' is defined, the last member will
// be copied into the stack of 'RScheckSuperframe' every time the function is
// called. Doing so, the compiler threats it as 'local'. However, the time
// inside of 'RScheckSuperframe' is very short and copying 256 bytes at entry
// takes some time too, so it's likely that you don't see any difference.

#define ATO_MOD_SIZE 512 // must be 512 or 768
//#define LOCAL_COPY

struct RS_LookUp {
    unsigned char RS_ato_mod[ATO_MOD_SIZE];
    unsigned char RS_iof[256];
};


const char 
szDllName[] = "viterbi",
szExtDLL[]  = ".dll",
szExtTXT[]  = ".txt",
szSSE2[]    = "SSE2",
szSSSE3[]   = "SSSE3",
szSSE41[]   = "SSE4.1",
szAVX[]     = "AVX",
szAVX2[]    = "AVX2",
szAVX5[]    = "AVX512";



struct DispatcherCMD {
    int InstruSet;  // wanted instruction-set (0-5)
    int ShowIRect;  // show info-rect on screen (0 or 1)
};


typedef int decon(unsigned int, unsigned int*, int, unsigned char*);
typedef void wakeupymm();
